from typing import Final

from faran.types import (
    NumPyNoiseCovariances,
    NumPyNoiseCovarianceDescription,
    JaxNoiseCovariances,
    JaxNoiseCovarianceDescription,
)
from faran.filters.kf import numpy_kalman_filter, jax_kalman_filter
from faran.filters.noise import (
    NumPyAdaptiveNoiseProvider,
    NumPyClampedNoiseProvider,
    NumPyNoiseCovarianceBounds,
    JaxAdaptiveNoiseProvider,
    JaxClampedNoiseProvider,
    JaxNoiseCovarianceBounds,
    IdentityNoiseModelProvider,
)


class noise:
    """Factory namespace for noise model providers."""

    class numpy:
        adaptive: Final = NumPyAdaptiveNoiseProvider.create
        clamped: Final = NumPyClampedNoiseProvider.decorate
        identity: Final = IdentityNoiseModelProvider

        @staticmethod
        def covariances(
            *,
            process: NumPyNoiseCovarianceDescription,
            observation: NumPyNoiseCovarianceDescription,
            process_dimension: int | None = None,
            observation_dimension: int | None = None,
        ) -> NumPyNoiseCovariances:
            return NumPyNoiseCovariances(
                process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                    process, dimension=process_dimension
                ),
                observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                    observation, dimension=observation_dimension
                ),
            )

        covariance_bounds: Final = NumPyNoiseCovarianceBounds.create

    class jax:
        adaptive: Final = JaxAdaptiveNoiseProvider.create
        clamped: Final = JaxClampedNoiseProvider.decorate
        identity: Final = IdentityNoiseModelProvider

        @staticmethod
        def covariances(
            *,
            process: JaxNoiseCovarianceDescription,
            observation: JaxNoiseCovarianceDescription,
            process_dimension: int | None = None,
            observation_dimension: int | None = None,
        ) -> JaxNoiseCovariances:
            return JaxNoiseCovariances(
                process_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                    process, dimension=process_dimension
                ),
                observation_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                    observation, dimension=observation_dimension
                ),
            )

        covariance_bounds: Final = JaxNoiseCovarianceBounds.create



from .common import (
    POSE_D_O as POSE_D_O,
    PoseD_o as PoseD_o,
    SampledObstacleStates as SampledObstacleStates,
    SampledObstaclePositions as SampledObstaclePositions,
    SampledObstacleHeadings as SampledObstacleHeadings,
    ObstacleStatesForTimeStep as ObstacleStatesForTimeStep,
    ObstacleStates as ObstacleStates,
    SampledObstaclePositionExtractor as SampledObstaclePositionExtractor,
    SampledObstacleHeadingExtractor as SampledObstacleHeadingExtractor,
    ObstacleStateProvider as ObstacleStateProvider,
    ObstacleStateSampler as ObstacleStateSampler,
    Distance as Distance,
    DistanceExtractor as DistanceExtractor,
    SampleCostFunction as SampleCostFunction,
    Risk as Risk,
    RiskMetric as RiskMetric,
)
from .basic import (
    NumPySampledObstacleStates as NumPySampledObstacleStates,
    NumPySampledObstaclePositions as NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings as NumPySampledObstacleHeadings,
    NumPyObstacleStatesForTimeStep as NumPyObstacleStatesForTimeStep,
    NumPyObstacleStates as NumPyObstacleStates,
    NumPySampledObstaclePositionExtractor as NumPySampledObstaclePositionExtractor,
    NumPySampledObstacleHeadingExtractor as NumPySampledObstacleHeadingExtractor,
    NumPyObstacleStateProvider as NumPyObstacleStateProvider,
    NumPyObstacleStateSampler as NumPyObstacleStateSampler,
    NumPyDistanceExtractor as NumPyDistanceExtractor,
    NumPyRisk as NumPyRisk,
    NumPyRiskMetric as NumPyRiskMetric,
)
from .accelerated import (
    JaxSampledObstacleStates as JaxSampledObstacleStates,
    JaxSampledObstaclePositions as JaxSampledObstaclePositions,
    JaxSampledObstacleHeadings as JaxSampledObstacleHeadings,
    JaxObstacleStatesForTimeStep as JaxObstacleStatesForTimeStep,
    JaxObstacleStates as JaxObstacleStates,
    JaxSampledObstaclePositionExtractor as JaxSampledObstaclePositionExtractor,
    JaxSampledObstacleHeadingExtractor as JaxSampledObstacleHeadingExtractor,
    JaxObstacleStateProvider as JaxObstacleStateProvider,
    JaxObstacleStateSampler as JaxObstacleStateSampler,
    JaxDistanceExtractor as JaxDistanceExtractor,
    JaxRisk as JaxRisk,
    JaxRiskMetric as JaxRiskMetric,
)



from typing import NamedTuple

from faran.types import (
    jaxtyped,
    JaxGaussianBelief,
    JaxNoiseCovariances,
    JaxNoiseModelProvider,
    JaxNoiseCovarianceDescription,
)
from faran.filters.noise import IdentityNoiseModelProvider

from jaxtyping import Array as JaxArray, Float

import equinox as eqx
import jax
import jax.numpy as jnp


class ObstaclePartitioning(NamedTuple):
    should_update: Float[JaxArray, " K"]
    should_initialize: Float[JaxArray, " K"]


class JaxKalmanFilter(eqx.Module):
    """Kalman Filter for linear systems."""

    noise_model: JaxNoiseModelProvider

    @staticmethod
    def create(
        *, noise_model: JaxNoiseModelProvider | None = None
    ) -> "JaxKalmanFilter":
        """Creates a Kalman filter for linear systems.

        Args:
            noise_model: An optional noise model provider for adaptive noise estimation.
        """
        return JaxKalmanFilter(
            noise_model=IdentityNoiseModelProvider()
            if noise_model is None
            else noise_model
        )

    @eqx.filter_jit
    @jaxtyped
    def filter(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
        state_transition_matrix: Float[JaxArray, "D_x D_x"],
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        observation_matrix: Float[JaxArray, "D_z D_x"],
    ) -> JaxGaussianBelief:
        """Run the Kalman filter over a sequence of observations.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
            state_transition_matrix: A matrix for state prediction.
            process_noise_covariance: R matrix representing the covariance of process noise.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            observation_matrix: H matrix mapping state to observation space.
        """
        belief = self.initial_belief_from(
            observations,
            initial_state_covariance=initial_state_covariance,
        )
        noise = JaxNoiseCovariances(
            process_noise_covariance, observation_noise_covariance
        )
        adapt = self.noise_model(
            obstacle_count=observations.shape[2],
            observation_matrix=observation_matrix,
            noise=noise,
        )
        noise_state = adapt.state

        def step(carry, observation):
            belief, noise, noise_state = carry
            belief = self.predict(
                belief=belief,
                state_transition_matrix=state_transition_matrix,
                process_noise_covariance=noise.process_noise_covariance,
            )
            noise, noise_state = adapt(
                noise=noise,
                prediction=belief,
                observation=observation,
                state=noise_state,
            )
            belief = self.update(
                observation,
                prediction=belief,
                observation_matrix=observation_matrix,
                observation_noise_covariance=noise.observation_noise_covariance,
                initial_state_covariance=initial_state_covariance,
            )
            return (belief, noise, noise_state), None

        (belief, _, _), _ = jax.lax.scan(
            step, (belief, noise, noise_state), observations
        )
        return belief

    @eqx.filter_jit
    @jaxtyped
    def predict(
        self,
        *,
        belief: JaxGaussianBelief,
        state_transition_matrix: Float[JaxArray, "D_x D_x"],
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the prediction step of the Kalman filter from the existing belief.

        Args:
            belief: The current belief about the state.
            state_transition_matrix: A matrix for state prediction.
            process_noise_covariance: R matrix representing the covariance of process noise.
        """
        return jax_kalman_filter.predict(
            belief=belief,
            state_transition_matrix=state_transition_matrix,
            process_noise_covariance=process_noise_covariance,
        )

    @eqx.filter_jit
    @jaxtyped
    def update(
        self,
        observation: Float[JaxArray, "D_z K"],
        *,
        prediction: JaxGaussianBelief,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the update step of the Kalman filter using a new observation.

        Args:
            observation: The newly observed state.
            prediction: The predicted belief from the prediction step.
            observation_matrix: H matrix mapping state to observation space.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.update(
            observation,
            prediction=prediction,
            observation_matrix=observation_matrix,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
        )

    @eqx.filter_jit
    @jaxtyped
    def initial_belief_from(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Initializes the belief state from the first observation using a pseudo-inverse.

        This method provides a simple way to initialize the state mean based on the first
        observation, while using the provided initial state covariance for uncertainty.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.initial_belief_from(
            observations,
            initial_state_covariance=initial_state_covariance,
        )


class jax_kalman_filter:
    @staticmethod
    @jax.jit
    @jaxtyped
    def predict(
        *,
        belief: JaxGaussianBelief,
        state_transition_matrix: Float[JaxArray, "D_x D_x"],
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        A = state_transition_matrix
        R = process_noise_covariance
        mu, sigma = belief

        return JaxGaussianBelief(
            mean=A @ mu,
            covariance=jnp.einsum("ij,jlk,ml->imk", A, sigma, A) + R[..., jnp.newaxis],
        )

    @staticmethod
    @jax.jit
    @jaxtyped
    def update(
        observation: Float[JaxArray, "D_z K"],
        *,
        prediction: JaxGaussianBelief,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        H = observation_matrix
        Q = observation_noise_covariance

        def partition(
            *, observation: Float[JaxArray, "D_z K"], mean: Float[JaxArray, "D_x K"]
        ) -> ObstaclePartitioning:
            observation_valid = ~jnp.any(jnp.isnan(observation), axis=0)
            prediction_valid = ~jnp.any(jnp.isnan(mean), axis=0)

            return ObstaclePartitioning(
                should_update=observation_valid & prediction_valid,
                should_initialize=observation_valid & ~prediction_valid,
            )

        def substitute_missing_values(
            *,
            observation: Float[JaxArray, "D_z K"],
            mean: Float[JaxArray, "D_x K"],
            covariance: Float[JaxArray, "D_x D_x K"],
        ) -> tuple[
            Float[JaxArray, "D_z K"],
            Float[JaxArray, "D_x K"],
            Float[JaxArray, "D_x D_x K"],
        ]:
            covariance_is_nan = jnp.any(jnp.isnan(covariance), axis=(0, 1))

            return (
                jnp.where(jnp.isnan(observation), 0.0, observation),
                jnp.where(jnp.isnan(mean), 0.0, mean),
                jnp.where(
                    covariance_is_nan[jnp.newaxis, jnp.newaxis, :],
                    initial_state_covariance[..., jnp.newaxis],
                    covariance,
                ),
            )

        def update(
            *,
            observation: Float[JaxArray, "D_z K"],
            mean: Float[JaxArray, "D_x K"],
            covariance: Float[JaxArray, "D_x D_x K"],
        ) -> JaxGaussianBelief:
            D_x = prediction.mean.shape[0]

            S = jnp.einsum("ij,jlk,ml->imk", H, covariance, H) + Q[..., jnp.newaxis]
            rhs = jnp.einsum("ijk,lj->ilk", covariance, H)
            K = jnp.linalg.solve(
                S.transpose(2, 0, 1), rhs.transpose(2, 1, 0)
            ).transpose(2, 1, 0)

            innovation = observation - H @ mean
            IKH = jnp.eye(D_x)[..., jnp.newaxis] - jnp.einsum("ijk,jl->ilk", K, H)

            joseph_term = jnp.einsum(
                "ijk,jlk->ilk", IKH, jnp.einsum("ijk,ljk->ilk", covariance, IKH)
            )
            kalman_noise_term = jnp.einsum("ijk,jl,mlk->imk", K, Q, K)

            return JaxGaussianBelief(
                mean=mean + jnp.einsum("ijk,jk->ik", K, innovation),
                covariance=joseph_term + kalman_noise_term,
            )

        def initialize_from(
            observation: Float[JaxArray, "D_z K"],
        ) -> JaxGaussianBelief:
            K = observation.shape[1]

            # NOTE: The mean state is the observation when available, 0 otherwise.
            return JaxGaussianBelief(
                mean=jnp.linalg.pinv(H) @ observation,
                covariance=jnp.repeat(
                    initial_state_covariance[..., jnp.newaxis], K, axis=2
                ),
            )

        def blend(
            partitioning: ObstaclePartitioning,
            *,
            initial: JaxGaussianBelief,
            prediction: JaxGaussianBelief,
            update: JaxGaussianBelief,
        ) -> JaxGaussianBelief:
            return JaxGaussianBelief(
                mean=jnp.select(
                    [partitioning.should_update, partitioning.should_initialize],
                    [update.mean, initial.mean],
                    default=prediction.mean,
                ),
                covariance=jnp.select(
                    [partitioning.should_update, partitioning.should_initialize],
                    [update.covariance, initial.covariance],
                    default=prediction.covariance,
                ),
            )

        partitioning = partition(observation=observation, mean=prediction.mean)
        observation, mean, covariance = substitute_missing_values(
            observation=observation,
            mean=prediction.mean,
            covariance=prediction.covariance,
        )

        return blend(
            partitioning,
            initial=initialize_from(observation),
            prediction=prediction,
            update=update(observation=observation, mean=mean, covariance=covariance),
        )

    @staticmethod
    @jax.jit
    @jaxtyped
    def initial_belief_from(
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        D_x = initial_state_covariance.shape[0]
        K = observations.shape[2]

        return JaxGaussianBelief(
            mean=jnp.full((D_x, K), jnp.nan),
            covariance=jnp.full((D_x, D_x, K), jnp.nan),
        )

    @staticmethod
    def standardize_noise_covariance(
        covariance: JaxNoiseCovarianceDescription, *, dimension: int | None = None
    ) -> Float[JaxArray, "D_c D_c"]:
        if isinstance(covariance, (int, float)):
            assert dimension is not None, (
                "Dimension must be provided for scalar covariance."
            )
            return covariance * jnp.eye(dimension)

        covariance = jnp.asarray(covariance)

        match covariance.ndim:
            case 2:
                assert dimension is None or covariance.shape == (
                    dimension,
                    dimension,
                ), (
                    f"Expected covariance shape ({dimension}, {dimension}), got {covariance.shape}."
                )
                return covariance
            case 1:
                assert dimension is None or covariance.shape == (dimension,), (
                    f"Expected covariance shape ({dimension},), got {covariance.shape}."
                )
                return jnp.diag(covariance)
            case _:
                assert False, f"Invalid covariance shape: {covariance.shape}"



from typing import Protocol, Self, Any
from dataclasses import dataclass

from faran.types.array import Array, DataType

from jaxtyping import Float, Int


@dataclass(kw_only=True, frozen=True)
class EstimatedObstacleStates[StatesT, InputsT, CovarianceT]:
    states: StatesT
    inputs: InputsT
    covariance: CovarianceT | None


class ObstacleStatesHistory[ObstacleStatesForTimeStepT = Any](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K"]:
        """Returns the obstacle history as a NumPy array."""
        ...

    def last(self) -> ObstacleStatesForTimeStepT:
        """Returns the obstacle states at the last time step."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon of the obstacle history."""
        ...

    @property
    def dimension(self) -> int:
        """The dimension of the obstacle states."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...


class ObstacleIds(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Int[Array, " K"]:
        """Returns the obstacle IDs as a NumPy array."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...


class ObstacleIdAssignment[ObstacleStatesForTimeStepT, IdT, HistoryT](Protocol):
    def __call__(
        self, states: ObstacleStatesForTimeStepT, /, *, history: HistoryT, ids: IdT
    ) -> IdT | None:
        """Returns the IDs to be assigned to the given obstacle states. If no IDs are returned,
        the obstacle states are assumed to be already ordered by their IDs.

        Args:
            states: The obstacle states at the current time step.
            history: The full history of obstacle states.
            ids: The obstacle IDs corresponding to the historical states.
        """
        ...


class ObstacleStatesRunningHistory[ObstacleStatesForTimeStepT, IdT, HistoryT](Protocol):
    def get(self) -> HistoryT:
        """Returns the current full running history of obstacle states."""
        ...

    def ids(self) -> IdT:
        """Returns the current obstacle IDs corresponding to the running history."""
        ...

    def append(self, states: ObstacleStatesForTimeStepT, /, *, ids: IdT | None) -> Self:
        """Appends new obstacle states at a time step to the running history. If the IDs are
        provided, the states must be reordered to match the IDs before appending.
        """
        ...


class ObstacleStateEstimator[HistoryT, StatesT, InputsT, CovarianceT = None](Protocol):
    def estimate_from(
        self, history: HistoryT
    ) -> EstimatedObstacleStates[StatesT, InputsT, CovarianceT]:
        """Estimates the current states and inputs of obstacles given their history."""
        ...


class InputAssumptionProvider[InputsT](Protocol):
    def __call__(self, inputs: InputsT, /) -> InputsT:
        """Applies assumptions to the given inputs and returns the modified inputs."""
        ...


class ObstacleModel[HistoryT, StatesT, InputsT, CovariancesT, StateSequencesT](
    Protocol
):
    def forward(
        self,
        *,
        states: StatesT,
        inputs: InputsT,
        covariances: CovariancesT | None,
        horizon: int,
    ) -> StateSequencesT:
        """Simulates the objects forward in time given the current states, assumed inputs, and horizon.

        Args:
            states: The current states of obstacles.
            inputs: The assumed control inputs.
            covariances: Uncertainty associated with the current states and inputs (if available).
            horizon: The number of time steps to simulate.
        """
        ...


class PredictionCreator[StateSequencesT, PredictionT](Protocol):
    def __call__(self, *, states: StateSequencesT) -> PredictionT:
        """Creates a prediction from the given state sequences."""
        ...

    def empty(self, *, horizon: int) -> PredictionT:
        """Creates an empty obstacle state prediction with the specified horizon."""
        ...


class ObstacleMotionPredictor[HistoryT, PredictionT](Protocol):
    def predict(self, *, history: HistoryT) -> PredictionT:
        """Predicts the future states of obstacles based on their state history."""
        ...



from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    Distance,
    DistanceExtractor,
    StateSequence,
    ObstacleStates,
    SimulationData,
    Metric,
)
from faran.collectors import access

from jaxtyping import Float, Bool

import numpy as np


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class CollisionMetricResult:
    """Results of the collision metric, including distances and collision flags.

    Attributes:
        distances: Distance between the ego vehicle and each obstacle at each time step.
        distance_threshold: Distance at or below which a collision is registered.
    """

    distances: Float[Array, "T V"]
    distance_threshold: float

    @cached_property
    def min_distances(self) -> Float[Array, " V"]:
        return self.distances.min(axis=0)

    @cached_property
    def collisions(self) -> Bool[Array, "T V"]:
        return self.distances <= self.distance_threshold

    @cached_property
    def collision_detected(self) -> bool:
        return bool(self.collisions.any())


@dataclass(kw_only=True, frozen=True)
class CollisionMetric[StateBatchT, SampledObstacleStatesT](
    Metric[CollisionMetricResult]
):
    """Metric evaluating minimum distances between the ego vehicle and obstacles."""

    distance: DistanceExtractor[StateBatchT, SampledObstacleStatesT, Distance]
    distance_threshold: float

    @staticmethod
    def create[S, SOS](
        *,
        distance_threshold: float,
        distance: DistanceExtractor[S, SOS, Distance],
    ) -> "CollisionMetric":
        return CollisionMetric(distance=distance, distance_threshold=distance_threshold)

    def compute(self, data: SimulationData) -> CollisionMetricResult:
        states = data(access.states.assume(StateSequence[StateBatchT]).require())
        obstacle_states = data(
            access.obstacle_states.assume(
                ObstacleStates[SampledObstacleStatesT]
            ).require()
        )

        measured_distances = self.distance(
            states=states.batched(), obstacle_states=obstacle_states.single()
        )

        distances = np.asarray(measured_distances).reshape(states.horizon, -1)

        return CollisionMetricResult(
            distances=distances, distance_threshold=self.distance_threshold
        )

    @property
    def name(self) -> str:
        return "collision"



from .common import ConvexPolygon as ConvexPolygon
from .basic import NumPySatDistanceExtractor as NumPySatDistanceExtractor
from .accelerated import JaxSatDistanceExtractor as JaxSatDistanceExtractor



from .common import (
    UNICYCLE_D_X as UNICYCLE_D_X,
    UNICYCLE_D_U as UNICYCLE_D_U,
    UNICYCLE_D_O as UNICYCLE_D_O,
    UnicycleD_x as UnicycleD_x,
    UnicycleD_u as UnicycleD_u,
    UnicycleD_o as UnicycleD_o,
    UnicycleState as UnicycleState,
    UnicycleStateSequence as UnicycleStateSequence,
    UnicycleStateBatch as UnicycleStateBatch,
    UnicyclePositions as UnicyclePositions,
    UnicycleControlInputSequence as UnicycleControlInputSequence,
    UnicycleControlInputBatch as UnicycleControlInputBatch,
)
from .basic import (
    NumPyUnicycleObstacleStatesHistory as NumPyUnicycleObstacleStatesHistory,
)
from .accelerated import (
    JaxUnicycleObstacleStatesHistory as JaxUnicycleObstacleStatesHistory,
)



from .basic import (
    NumPyHungarianObstacleIdAssignment as NumPyHungarianObstacleIdAssignment,
)
from .accelerated import (
    JaxHungarianObstacleIdAssignment as JaxHungarianObstacleIdAssignment,
)



from typing import Self, Protocol
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    DataType,
    Array,
    jaxtyped,
    NumPyObstacleStatesForTimeStep,
    JaxObstacleStates,
    JaxObstacleStatesForTimeStep,
)
from faran.obstacles.history.basic import (
    NumPyObstacleIds,
    NumPyObstacleStatesRunningHistory,
)

from jaxtyping import Array as JaxArray, Float, Int, Num

import numpy as np
import jax.numpy as jnp


class JaxObstacleStateCreator[StatesT](Protocol):
    """Protocol for creating obstacle state objects from JAX arrays."""

    def wrap(self, states: Float[JaxArray, "T D_o K"], /) -> StatesT:
        """Wraps a JAX array into the appropriate obstacle states type."""
        ...

    def empty(self, *, horizon: int, obstacle_count: int) -> StatesT:
        """Creates empty obstacle states with the given horizon and obstacle count."""
        ...


@dataclass(frozen=True)
class JaxObstacleStateCreatorAdapter[StatesT]:
    """Adapts a JAX obstacle state creator to accept NumPy arrays."""

    _creator: JaxObstacleStateCreator[StatesT]

    def wrap(self, states: Float[Array, "T D_o K"]) -> StatesT:
        return self._creator.wrap(jnp.asarray(states))

    def empty(self, *, horizon: int, obstacle_count: int) -> StatesT:
        return self._creator.empty(horizon=horizon, obstacle_count=obstacle_count)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacleIds:
    """JAX container for integer obstacle identifiers."""

    _ids: Int[JaxArray, " K"]

    @staticmethod
    def create(*, ids: Num[Array, " K"] | Num[JaxArray, " K"]) -> "JaxObstacleIds":
        return JaxObstacleIds(_ids=jnp.asarray(ids).astype(jnp.int32))

    def __array__(self, dtype: DataType | None = None) -> Int[Array, " K"]:
        return self._numpy_array

    def numpy(self) -> NumPyObstacleIds:
        return NumPyObstacleIds.create(ids=self._numpy_array)

    @property
    def count(self) -> int:
        return self._ids.shape[0]

    @property
    def array(self) -> Int[JaxArray, " K"]:
        return self._ids

    @property
    def _numpy_array(self) -> Int[Array, " K"]:
        return np.asarray(self._ids, dtype=np.intp)


@dataclass(kw_only=True, frozen=True)
class JaxObstacleStatesRunningHistory[
    StatesT: JaxObstacleStates,
    StatesForTimeStepT: JaxObstacleStatesForTimeStep,
]:
    """JAX sliding window history of obstacle states, delegating to NumPy internally."""

    # NOTE: Internally uses NumPy implementation, since it's faster in most cases.
    history: NumPyObstacleStatesRunningHistory[StatesT, NumPyObstacleStatesForTimeStep]

    @staticmethod
    def empty[
        S: JaxObstacleStates,
        STS: JaxObstacleStatesForTimeStep = JaxObstacleStatesForTimeStep,
    ](
        *,
        creator: JaxObstacleStateCreator[S],
        horizon: int | None = None,
        obstacle_count: int | None = None,
    ) -> "JaxObstacleStatesRunningHistory[S, STS]":
        return JaxObstacleStatesRunningHistory(
            history=NumPyObstacleStatesRunningHistory.empty(
                creator=JaxObstacleStateCreatorAdapter(creator),
                horizon=horizon,
                obstacle_count=obstacle_count,
            )
        )

    @staticmethod
    def single[S: JaxObstacleStates, STS: JaxObstacleStatesForTimeStep](
        observation: STS,
        *,
        creator: JaxObstacleStateCreator[S],
        horizon: int | None = None,
        obstacle_count: int | None = None,
    ) -> "JaxObstacleStatesRunningHistory[S, STS]":
        return JaxObstacleStatesRunningHistory(
            history=NumPyObstacleStatesRunningHistory.single(
                observation=observation.numpy(),
                creator=JaxObstacleStateCreatorAdapter(creator),
                horizon=horizon,
                obstacle_count=obstacle_count,
            )
        )

    def last(self) -> StatesForTimeStepT:
        return self._get.last()

    def get(self) -> StatesT:
        return self._get

    def ids(self) -> JaxObstacleIds:
        return self._ids

    def append(
        self, observation: StatesForTimeStepT, *, ids: JaxObstacleIds | None = None
    ) -> Self:
        return self.__class__(
            history=self.history.append(
                observation=observation.numpy(),
                ids=ids.numpy() if ids is not None else None,
            )
        )

    @property
    def horizon(self) -> int:
        return self.history.horizon

    @property
    def count(self) -> int:
        return self.history.count

    @cached_property
    def _get(self) -> StatesT:
        return self.history.get()

    @cached_property
    def _ids(self) -> JaxObstacleIds:
        return JaxObstacleIds.create(ids=jnp.asarray(self.history.ids().array))



from .basic import (
    NumPyContouringCost as NumPyContouringCost,
    NumPyLagCost as NumPyLagCost,
    NumPyProgressCost as NumPyProgressCost,
    NumPyControlSmoothingCost as NumPyControlSmoothingCost,
    NumPyControlEffortCost as NumPyControlEffortCost,
)
from .accelerated import (
    JaxContouringCost as JaxContouringCost,
    JaxLagCost as JaxLagCost,
    JaxProgressCost as JaxProgressCost,
    JaxControlSmoothingCost as JaxControlSmoothingCost,
    JaxControlEffortCost as JaxControlEffortCost,
)
from .combined import (
    CombinedCost as CombinedCost,
    NumPyCostSumFunction as NumPyCostSumFunction,
    JaxCostSumFunction as JaxCostSumFunction,
)
from .collision import (
    NumPyDistance as NumPyDistance,
    NumPyCollisionCost as NumPyCollisionCost,
    JaxDistance as JaxDistance,
    JaxCollisionCost as JaxCollisionCost,
)
from .distance import (
    Circles as Circles,
    ConvexPolygon as ConvexPolygon,
    NumPyCircleDistanceExtractor as NumPyCircleDistanceExtractor,
    JaxCircleDistanceExtractor as JaxCircleDistanceExtractor,
    NumPySatDistanceExtractor as NumPySatDistanceExtractor,
    JaxSatDistanceExtractor as JaxSatDistanceExtractor,
)
from .boundary import (
    NumPyFixedWidthBoundary as NumPyFixedWidthBoundary,
    JaxFixedWidthBoundary as JaxFixedWidthBoundary,
)
from .risk import risk as risk
from .factory import costs as costs, distance as distance, boundary as boundary



from typing import Protocol
from dataclasses import dataclass
from functools import cached_property

from faran.types.array import Array, DataType, jaxtyped
from faran.types.costs.boundary.common import BoundaryDistanceExtractor

from jaxtyping import Array as JaxArray, Float

import numpy as np
import jax.numpy as jnp


class JaxBoundaryDistanceExtractor[StateBatchT, DistanceT](
    BoundaryDistanceExtractor[StateBatchT, DistanceT], Protocol
): ...


@jaxtyped
@dataclass(frozen=True)
class JaxBoundaryDistance:
    _array: Float[JaxArray, "T M"]

    @staticmethod
    def create(
        *, array: Float[Array, "T M"] | Float[JaxArray, "T M"]
    ) -> "JaxBoundaryDistance":
        """Creates a JAX boundary distance from the given array."""
        return JaxBoundaryDistance(jnp.asarray(array))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T M"]:
        return np.asarray(self._array)



from typing import NamedTuple

from faran.types import (
    jaxtyped,
    JaxNoiseModel,
    JaxNoiseModelProvider,
    JaxGaussianBelief,
    JaxNoiseCovariances,
)

from jaxtyping import Array as JaxArray, Bool, Float, Int, Scalar

import equinox as eqx
import jax
import jax.numpy as jnp


class JaxNoiseCovarianceBounds(NamedTuple):
    process: Scalar
    observation: Scalar

    @staticmethod
    def create(*, process: float, observation: float) -> "JaxNoiseCovarianceBounds":
        """Creates isotropic bounds for noise covariances.

        Args:
            process: Isotropic bound for process noise covariance eigenvalues.
            observation: Isotropic bound for observation noise covariance eigenvalues.
        """
        return JaxNoiseCovarianceBounds(
            process=jnp.asarray(process), observation=jnp.asarray(observation)
        )


class JaxClampedNoise[StateT](eqx.Module):
    """Decorator that clamps an inner noise model's eigenvalues to a floor and/or ceiling."""

    inner: JaxNoiseModel[StateT]
    floor: JaxNoiseCovarianceBounds
    ceiling: JaxNoiseCovarianceBounds

    @eqx.filter_jit
    @jaxtyped
    def __call__(
        self,
        *,
        noise: JaxNoiseCovariances,
        prediction: JaxGaussianBelief,
        observation: Float[JaxArray, "D_z K"],
        state: StateT,
    ) -> tuple[JaxNoiseCovariances, StateT]:
        result, state = self.inner(
            noise=noise, prediction=prediction, observation=observation, state=state
        )
        return JaxNoiseCovariances(
            process_noise_covariance=clamp_eigenvalues(
                result.process_noise_covariance,
                floor=self.floor.process,
                ceiling=self.ceiling.process,
            ),
            observation_noise_covariance=clamp_eigenvalues(
                result.observation_noise_covariance,
                floor=self.floor.observation,
                ceiling=self.ceiling.observation,
            ),
        ), state

    @property
    def state(self) -> StateT:
        return self.inner.state


class JaxClampedNoiseProvider[StateT](eqx.Module):
    inner: JaxNoiseModelProvider[StateT]
    floor: JaxNoiseCovarianceBounds
    ceiling: JaxNoiseCovarianceBounds

    @staticmethod
    def decorate[S](
        inner: JaxNoiseModelProvider[S],
        *,
        floor: JaxNoiseCovarianceBounds | None = None,
        ceiling: JaxNoiseCovarianceBounds | None = None,
    ) -> "JaxClampedNoiseProvider[S]":
        """Creates a noise model provider that clamps the eigenvalues of the
        noise covariances to the specified floor and/or ceiling.

        Args:
            inner: The inner noise model provider to delegate to.
            floor: Isotropic minimum bounds. Eigenvalues of the inner model's
                output will be clamped to be no smaller than these.
            ceiling: Isotropic maximum bounds. Eigenvalues of the inner model's
                output will be clamped to be no larger than these.
        """
        return JaxClampedNoiseProvider(
            inner=inner,
            floor=floor
            or JaxNoiseCovarianceBounds(
                process=jnp.asarray(0.0), observation=jnp.asarray(0.0)
            ),
            ceiling=ceiling
            or JaxNoiseCovarianceBounds(
                process=jnp.asarray(jnp.inf), observation=jnp.asarray(jnp.inf)
            ),
        )

    @eqx.filter_jit
    def __call__(
        self,
        *,
        obstacle_count: int,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        noise: JaxNoiseCovariances,
    ) -> JaxClampedNoise:
        inner_model = self.inner(
            obstacle_count=obstacle_count,
            observation_matrix=observation_matrix,
            noise=noise,
        )
        return JaxClampedNoise(
            inner=inner_model, floor=self.floor, ceiling=self.ceiling
        )


class JaxAdaptiveNoiseState(NamedTuple):
    """Circular buffer state for adaptive noise estimation."""

    buffer: Float[JaxArray, "W D_z D_z K"]
    entry_count: Int[JaxArray, " K"]


class JaxAdaptiveNoise(eqx.Module):
    """Innovation-Based Adaptive Estimation (IAE) for noise covariances."""

    observation_matrix: Float[JaxArray, "D_z D_x"]
    obstacle_count: int = eqx.field(static=True)
    window_size: int = eqx.field(static=True)

    @eqx.filter_jit
    @jaxtyped
    def __call__(
        self,
        *,
        noise: JaxNoiseCovariances,
        prediction: JaxGaussianBelief,
        observation: Float[JaxArray, "D_z K"],
        state: JaxAdaptiveNoiseState,
    ) -> tuple[JaxNoiseCovariances, JaxAdaptiveNoiseState]:
        valid = valid_obstacle_mask(prediction, observation)
        has_valid = jnp.any(valid)

        return jax.lax.cond(
            has_valid,
            lambda _: self.adapt(
                noise=noise,
                prediction=prediction,
                observation=observation,
                state=state,
                valid=valid,
            ),
            lambda _: (noise, state),
            None,
        )

    @eqx.filter_jit
    @jaxtyped
    def adapt(
        self,
        *,
        noise: JaxNoiseCovariances,
        prediction: JaxGaussianBelief,
        observation: Float[JaxArray, "D_z K"],
        state: JaxAdaptiveNoiseState,
        valid: Bool[JaxArray, " K"],
    ) -> tuple[JaxNoiseCovariances, JaxAdaptiveNoiseState]:
        innovation = compute_innovation_matrix(
            prediction=prediction,
            observation=observation,
            observation_matrix=self.observation_matrix,
        )
        new_state = compute_updated_state(
            state=state, innovation=innovation, valid=valid
        )

        any_buffer_full = jnp.any(
            buffer_full := new_state.entry_count >= self.window_size
        )

        def return_original(
            _: None,
        ) -> tuple[JaxNoiseCovariances, JaxAdaptiveNoiseState]:
            return noise, new_state

        def compute_adapted(
            _: None,
        ) -> tuple[JaxNoiseCovariances, JaxAdaptiveNoiseState]:
            innovation_matrices = jnp.median(new_state.buffer, axis=0)
            safe_covariance = jnp.where(
                jnp.isnan(prediction.covariance), 0.0, prediction.covariance
            )

            kalman_gains = jax.vmap(
                lambda cov: compute_kalman_gain(
                    predicted_covariance=cov,
                    observation_matrix=self.observation_matrix,
                    observation_noise_covariance=noise.observation_noise_covariance,
                ),
                in_axes=2,
                out_axes=2,
            )(safe_covariance)

            adapted_process = jax.vmap(
                lambda K, V: enforce_spd(K @ V @ K.T),
                in_axes=(2, 2),
                out_axes=2,
            )(kalman_gains, innovation_matrices)

            adapted_observation = jax.vmap(
                lambda V, P: enforce_spd(
                    V - self.observation_matrix @ P @ self.observation_matrix.T
                ),
                in_axes=(2, 2),
                out_axes=2,
            )(innovation_matrices, safe_covariance)

            adapted_noise = aggregate_noise_covariances(
                buffer_full=buffer_full,
                valid=valid,
                adapted_process=adapted_process,
                adapted_observation=adapted_observation,
                noise=noise,
            )

            return adapted_noise, new_state

        return jax.lax.cond(any_buffer_full, compute_adapted, return_original, None)

    @property
    def state(self) -> JaxAdaptiveNoiseState:
        observation_dimension = self.observation_matrix.shape[0]
        return JaxAdaptiveNoiseState(
            buffer=jnp.zeros(
                (
                    self.window_size,
                    observation_dimension,
                    observation_dimension,
                    self.obstacle_count,
                )
            ),
            entry_count=jnp.zeros(self.obstacle_count, dtype=jnp.int32),
        )


class JaxAdaptiveNoiseProvider(eqx.Module):
    window_size: int = eqx.field(static=True)

    @staticmethod
    def create(*, window_size: int) -> "JaxAdaptiveNoiseProvider":
        """Creates an innovation-based adaptive estimation model for noise.

        Args:
            window_size: The number of past observations considered, when adapting
                the noise covariances.
        """
        return JaxAdaptiveNoiseProvider(window_size=window_size)

    def __call__(
        self,
        *,
        obstacle_count: int,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        noise: JaxNoiseCovariances,
    ) -> JaxAdaptiveNoise:
        return JaxAdaptiveNoise(
            observation_matrix=observation_matrix,
            obstacle_count=obstacle_count,
            window_size=self.window_size,
        )


def valid_obstacle_mask(
    prediction: JaxGaussianBelief, observation: Float[JaxArray, "D_z K"]
) -> Bool[JaxArray, " K"]:
    mean_valid = ~jnp.any(jnp.isnan(prediction.mean), axis=0)
    covariance_valid = ~jnp.any(
        jnp.isnan(prediction.covariance.reshape(-1, prediction.covariance.shape[2])),
        axis=0,
    )
    observation_valid = ~jnp.any(jnp.isnan(observation), axis=0)
    return mean_valid & covariance_valid & observation_valid


@jax.jit
@jaxtyped
def compute_innovation_matrix(
    prediction: JaxGaussianBelief,
    observation: Float[JaxArray, "D_z K"],
    *,
    observation_matrix: Float[JaxArray, "D_z D_x"],
) -> Float[JaxArray, "D_z D_z K"]:
    safe_mean = jnp.where(jnp.isnan(prediction.mean), 0.0, prediction.mean)
    safe_observation = jnp.where(jnp.isnan(observation), 0.0, observation)
    innovation = safe_observation - observation_matrix @ safe_mean

    return innovation[:, jnp.newaxis, :] * innovation[jnp.newaxis, :, :]


@jax.jit
@jaxtyped
def compute_updated_state(
    state: JaxAdaptiveNoiseState,
    *,
    innovation: Float[JaxArray, "D_z D_z K"],
    valid: Bool[JaxArray, " K"],
) -> JaxAdaptiveNoiseState:
    window_size = state.buffer.shape[0]
    obstacle_count = valid.shape[0]

    indices = state.entry_count % window_size
    updated_buffer = state.buffer.at[indices, :, :, jnp.arange(obstacle_count)].set(
        innovation.transpose(2, 0, 1)
    )

    return JaxAdaptiveNoiseState(
        buffer=jnp.where(
            valid[jnp.newaxis, jnp.newaxis, jnp.newaxis, :],
            updated_buffer,
            state.buffer,
        ),
        entry_count=state.entry_count + valid.astype(jnp.int32),
    )


@jax.jit
@jaxtyped
def aggregate_noise_covariances(
    *,
    buffer_full: Bool[JaxArray, " K"],
    valid: Bool[JaxArray, " K"],
    adapted_process: Float[JaxArray, "D_x D_x K"],
    adapted_observation: Float[JaxArray, "D_z D_z K"],
    noise: JaxNoiseCovariances,
) -> JaxNoiseCovariances:
    use_adapted = buffer_full & valid
    valid_adapted_count = jnp.maximum(jnp.sum(use_adapted), 1)

    mean_process = (
        jnp.sum(
            jnp.where(
                use_adapted[jnp.newaxis, jnp.newaxis, :],
                adapted_process,
                0.0,
            ),
            axis=2,
        )
        / valid_adapted_count
    )

    mean_observation = (
        jnp.sum(
            jnp.where(
                use_adapted[jnp.newaxis, jnp.newaxis, :],
                adapted_observation,
                0.0,
            ),
            axis=2,
        )
        / valid_adapted_count
    )

    has_valid_adapted = jnp.any(use_adapted)

    return JaxNoiseCovariances(
        process_noise_covariance=jnp.where(
            has_valid_adapted,
            mean_process,
            noise.process_noise_covariance,
        ),
        observation_noise_covariance=jnp.where(
            has_valid_adapted,
            mean_observation,
            noise.observation_noise_covariance,
        ),
    )


@jax.jit
@jaxtyped
def compute_kalman_gain(
    *,
    predicted_covariance: Float[JaxArray, "D_x D_x"],
    observation_matrix: Float[JaxArray, "D_z D_x"],
    observation_noise_covariance: Float[JaxArray, "D_z D_z"],
) -> Float[JaxArray, "D_x D_z"]:
    S = (
        observation_matrix @ predicted_covariance @ observation_matrix.T
        + observation_noise_covariance
    )
    return jnp.linalg.solve(S, observation_matrix @ predicted_covariance).T


@jax.jit
@jaxtyped
def clamp_eigenvalues(
    matrix: Float[JaxArray, "N N"], *, floor: Scalar, ceiling: Scalar
) -> Float[JaxArray, "N N"]:
    symmetrised = (matrix + matrix.T) / 2
    eigenvalues, eigenvectors = jnp.linalg.eigh(symmetrised)
    clamped = jnp.clip(eigenvalues, floor, ceiling)
    return (eigenvectors * clamped[jnp.newaxis, :]) @ eigenvectors.T


@jax.jit
@jaxtyped
def enforce_spd(matrix: Float[JaxArray, "N N"]) -> Float[JaxArray, "N N"]:
    eps = 1e-8
    symmetrised = (matrix + matrix.T) / 2
    eigenvalues, eigenvectors = jnp.linalg.eigh(symmetrised)
    return eigenvectors @ jnp.diag(jnp.maximum(eigenvalues, eps)) @ eigenvectors.T



from .history import (
    ObstaclePositionsForTimeStep as ObstaclePositionsForTimeStep,
    ObstaclePositions as ObstaclePositions,
    ObstaclePositionExtractor as ObstaclePositionExtractor,
    ObstacleOrientationsForTimeStep as ObstacleOrientationsForTimeStep,
    ObstacleOrientations as ObstacleOrientations,
    ObstacleOrientationExtractor as ObstacleOrientationExtractor,
    NumPyObstaclePositionsForTimeStep as NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions as NumPyObstaclePositions,
    NumPyObstaclePositionExtractor as NumPyObstaclePositionExtractor,
    NumPyObstacleOrientationsForTimeStep as NumPyObstacleOrientationsForTimeStep,
    NumPyObstacleOrientations as NumPyObstacleOrientations,
    NumPyObstacleOrientationExtractor as NumPyObstacleOrientationExtractor,
    JaxObstaclePositionsForTimeStep as JaxObstaclePositionsForTimeStep,
    JaxObstaclePositions as JaxObstaclePositions,
    JaxObstaclePositionExtractor as JaxObstaclePositionExtractor,
    JaxObstacleOrientationsForTimeStep as JaxObstacleOrientationsForTimeStep,
    JaxObstacleOrientations as JaxObstacleOrientations,
    JaxObstacleOrientationExtractor as JaxObstacleOrientationExtractor,
)
from .common import (
    ObstacleSimulator as ObstacleSimulator,
)
from .basic import (
    NumPyObstacleSimulator as NumPyObstacleSimulator,
)
from .accelerated import (
    JaxObstacleSimulator as JaxObstacleSimulator,
)



from typing import Sequence
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    POSE_D_O as D_O,
    Device,
    place,
    JaxSampledObstacleStates,
    JaxSampledObstaclePositions,
    JaxSampledObstacleHeadings,
    JaxObstacleStates,
    JaxObstacleStatesForTimeStep,
    JaxObstaclePositions,
    JaxObstaclePositionsForTimeStep,
    JaxObstacleOrientations,
    JaxObstacleOrientationsForTimeStep,
)
from faran.obstacles.basic import NumPyObstacle2dPosesForTimeStep

from numtypes import D
from jaxtyping import Array as JaxArray, Float

import numpy as np
import jax.numpy as jnp

ObstacleCovarianceArray = Float[JaxArray, f"T {D_O} {D_O} K"]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxSampledObstacle2dPoses(JaxSampledObstacleStates):
    """Sampled 2D poses (x, y, heading) with shape (T, POSE_D_O, K, N)."""

    _x: Float[JaxArray, "T K N"]
    _y: Float[JaxArray, "T K N"]
    _heading: Float[JaxArray, "T K N"]

    @staticmethod
    def create(
        *,
        x: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
        y: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
        heading: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
    ) -> "JaxSampledObstacle2dPoses":
        return JaxSampledObstacle2dPoses(
            _x=jnp.asarray(x), _y=jnp.asarray(y), _heading=jnp.asarray(heading)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K N"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T K N"]:
        return np.asarray(self._x)

    def y(self) -> Float[Array, "T K N"]:
        return np.asarray(self._y)

    def heading(self) -> Float[Array, "T K N"]:
        return np.asarray(self._heading)

    def positions(self) -> JaxSampledObstaclePositions:
        return JaxSampledObstaclePositions.create(x=self._x, y=self._y)

    def headings(self) -> JaxSampledObstacleHeadings:
        return JaxSampledObstacleHeadings.create(heading=self._heading)

    def at(self, *, time_step: int, sample: int) -> "JaxObstacle2dPosesForTimeStep":
        return JaxObstacle2dPosesForTimeStep.create(
            x=self._x[time_step, :, sample],
            y=self._y[time_step, :, sample],
            heading=self._heading[time_step, :, sample],
        )

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @property
    def sample_count(self) -> int:
        return self._x.shape[2]

    @property
    def x_array(self) -> Float[JaxArray, "T K N"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, "T K N"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, "T K N"]:
        return self._heading

    @property
    def array(self) -> Float[JaxArray, f"T {D_O} K N"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, f"T {D_O} K N"]:
        return jnp.stack([self._x, self._y, self._heading], axis=1)

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_o K N"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPositions(JaxObstaclePositions):
    """2D positions (x, y) with shape (T, 2, K)."""

    _x: Float[JaxArray, "T K"]
    _y: Float[JaxArray, "T K"]

    @staticmethod
    def create(
        *, x: Float[JaxArray, "T K"], y: Float[JaxArray, "T K"]
    ) -> "JaxObstacle2dPositions":
        return JaxObstacle2dPositions(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 K"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T 2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "T 2 K"]:
        return jnp.stack([self._x, self._y], axis=1)

    @cached_property
    def _numpy_array(self) -> Float[Array, "T 2 K"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPositionsForTimeStep(JaxObstaclePositionsForTimeStep):
    """2D positions (x, y) for a single time step with shape (2, K)."""

    _x: Float[JaxArray, " K"]
    _y: Float[JaxArray, " K"]

    @staticmethod
    def create(
        *, x: Float[JaxArray, " K"], y: Float[JaxArray, " K"]
    ) -> "JaxObstacle2dPositionsForTimeStep":
        return JaxObstacle2dPositionsForTimeStep(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "2 K"]:
        return self._numpy_array

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> int:
        return self._x.shape[0]

    @property
    def array(self) -> Float[JaxArray, "2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "2 K"]:
        return jnp.stack([self._x, self._y], axis=0)

    @cached_property
    def _numpy_array(self) -> Float[Array, "2 K"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacleHeadings(JaxObstacleOrientations):
    """Obstacle headings with shape (T, 1, K)."""

    _heading: Float[JaxArray, "T K"]

    @staticmethod
    def create(*, heading: Float[JaxArray, "T K"]) -> "JaxObstacleHeadings":
        return JaxObstacleHeadings(_heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 1 K"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self._heading.shape[0]

    @property
    def dimension(self) -> D[1]:
        return 1

    @property
    def count(self) -> int:
        return self._heading.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T 1 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "T 1 K"]:
        return self._heading[:, jnp.newaxis, :]

    @cached_property
    def _numpy_array(self) -> Float[Array, "T 1 K"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacleHeadingsForTimeStep(JaxObstacleOrientationsForTimeStep):
    """Obstacle headings for a single time step with shape (1, K)."""

    _heading: Float[JaxArray, " K"]

    @staticmethod
    def create(*, heading: Float[JaxArray, " K"]) -> "JaxObstacleHeadingsForTimeStep":
        return JaxObstacleHeadingsForTimeStep(_heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "1 K"]:
        return self._numpy_array

    @property
    def dimension(self) -> D[1]:
        return 1

    @property
    def count(self) -> int:
        return self._heading.shape[0]

    @property
    def array(self) -> Float[JaxArray, "1 K"]:
        return self._jax_array

    @cached_property
    def _jax_array(self) -> Float[JaxArray, "1 K"]:
        return self._heading[jnp.newaxis, :]

    @cached_property
    def _numpy_array(self) -> Float[Array, "1 K"]:
        return np.asarray(self._jax_array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPoses(
    JaxObstacleStates[
        JaxSampledObstacle2dPoses,
        "JaxObstacle2dPosesForTimeStep",
    ]
):
    """2D poses (x, y, heading) with shape (T, POSE_D_O, K)."""

    _x: Float[JaxArray, "T K"]
    _y: Float[JaxArray, "T K"]
    _heading: Float[JaxArray, "T K"]
    _covariance: ObstacleCovarianceArray | None = None

    @staticmethod
    def empty(*, horizon: int, obstacle_count: int = 0) -> "JaxObstacle2dPoses":
        """Creates obstacle states for zero obstacles over the given time horizon."""
        empty = jnp.full((horizon, obstacle_count), fill_value=jnp.nan)

        return JaxObstacle2dPoses.create(x=empty, y=empty, heading=empty)

    @staticmethod
    def sampled(  # type: ignore
        *,
        x: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
        y: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
        heading: Float[Array, "T K N"] | Float[JaxArray, "T K N"],
    ) -> JaxSampledObstacle2dPoses:
        return JaxSampledObstacle2dPoses.create(
            x=jnp.asarray(x), y=jnp.asarray(y), heading=jnp.asarray(heading)
        )

    @staticmethod
    def wrap(array: Float[JaxArray, f"T {D_O} K"]) -> "JaxObstacle2dPoses":
        return JaxObstacle2dPoses.create(
            x=array[:, 0, :], y=array[:, 1, :], heading=array[:, 2, :]
        )

    @staticmethod
    def create(
        *,
        x: Float[Array, "T K"] | Float[JaxArray, "T K"],
        y: Float[Array, "T K"] | Float[JaxArray, "T K"],
        heading: Float[Array, "T K"] | Float[JaxArray, "T K"],
        covariance: Float[Array, "T D_o D_o K"] | ObstacleCovarianceArray | None = None,
    ) -> "JaxObstacle2dPoses":
        return JaxObstacle2dPoses(
            _x=jnp.asarray(x),
            _y=jnp.asarray(y),
            _heading=jnp.asarray(heading),
            _covariance=jnp.asarray(covariance) if covariance is not None else None,
        )

    @staticmethod
    def of_states(
        obstacle_states: Sequence["JaxObstacle2dPosesForTimeStep"],
    ) -> "JaxObstacle2dPoses":
        assert len(obstacle_states) > 0, "Obstacle states sequence must not be empty."

        K = max(states.count for states in obstacle_states)

        def pad(array: Float[JaxArray, " L"]) -> Float[JaxArray, " K"]:
            return jnp.pad(array, (0, K - len(array)), constant_values=jnp.nan)

        x = jnp.stack([pad(states.x_array) for states in obstacle_states])
        y = jnp.stack([pad(states.y_array) for states in obstacle_states])
        heading = jnp.stack([pad(states.heading_array) for states in obstacle_states])

        return JaxObstacle2dPoses.create(x=x, y=y, heading=heading)

    @staticmethod
    def for_time_step(
        *,
        x: Float[Array, " K"] | Float[JaxArray, " K"],
        y: Float[Array, " K"] | Float[JaxArray, " K"],
        heading: Float[Array, " K"] | Float[JaxArray, " K"],
        device: Device = "cpu",
    ) -> "JaxObstacle2dPosesForTimeStep":
        """Creates obstacle states for a single time step.

        Note:
            Since the common case is to further process this data on the CPU first,
            the default device is set to "cpu".
        """
        return JaxObstacle2dPosesForTimeStep.create(
            x=place(x, device=device),
            y=place(y, device=device),
            heading=place(heading, device=device),
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T K"]:
        return np.asarray(self._x)

    def y(self) -> Float[Array, "T K"]:
        return np.asarray(self._y)

    def heading(self) -> Float[Array, "T K"]:
        return np.asarray(self._heading)

    def covariance(self) -> Float[Array, "T D_o D_o K"] | None:
        return np.asarray(self._covariance) if self._covariance is not None else None

    def positions(self) -> JaxObstacle2dPositions:
        return JaxObstacle2dPositions.create(x=self._x, y=self._y)

    def headings(self) -> JaxObstacleHeadings:
        return JaxObstacleHeadings.create(heading=self._heading)

    def single(self) -> JaxSampledObstacle2dPoses:
        return JaxSampledObstacle2dPoses.create(
            x=self._x[..., jnp.newaxis],
            y=self._y[..., jnp.newaxis],
            heading=self._heading[..., jnp.newaxis],
        )

    def last(self) -> "JaxObstacle2dPosesForTimeStep":
        return self.at(time_step=self.horizon - 1)

    def at(self, time_step: int) -> "JaxObstacle2dPosesForTimeStep":
        return JaxObstacle2dPosesForTimeStep.create(
            x=self._x[time_step],
            y=self._y[time_step],
            heading=self._heading[time_step],
        )

    @property
    def array(self) -> Float[JaxArray, "T D_o K"]:
        return self._array

    @property
    def x_array(self) -> Float[JaxArray, "T K"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, "T K"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, "T K"]:
        return self._heading

    @property
    def covariance_array(self) -> ObstacleCovarianceArray | None:
        return self._covariance

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @cached_property
    def _array(self) -> Float[JaxArray, "T D_o K"]:
        return jnp.stack([self._x, self._y, self._heading], axis=1)

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_o K"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPosesForTimeStep(JaxObstacleStatesForTimeStep[JaxObstacle2dPoses]):
    """2D poses (x, y, heading) for a single time step."""

    _x: Float[JaxArray, " K"]
    _y: Float[JaxArray, " K"]
    _heading: Float[JaxArray, " K"]

    @staticmethod
    def wrap(
        array: Float[Array, f"{D_O} K"] | Float[JaxArray, f"{D_O} K"],
    ) -> "JaxObstacle2dPosesForTimeStep":
        return JaxObstacle2dPosesForTimeStep.create(
            x=array[0, :], y=array[1, :], heading=array[2, :]
        )

    @staticmethod
    def create(
        *,
        x: Float[Array, " K"] | Float[JaxArray, " K"],
        y: Float[Array, " K"] | Float[JaxArray, " K"],
        heading: Float[Array, " K"] | Float[JaxArray, " K"],
    ) -> "JaxObstacle2dPosesForTimeStep":

        return JaxObstacle2dPosesForTimeStep(
            _x=jnp.asarray(x), _y=jnp.asarray(y), _heading=jnp.asarray(heading)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "D_o K"]:
        return self._numpy_array

    def numpy(self) -> NumPyObstacle2dPosesForTimeStep:
        return NumPyObstacle2dPosesForTimeStep.create(
            x=np.asarray(self._x),
            y=np.asarray(self._y),
            heading=np.asarray(self._heading),
        )

    def x(self) -> Float[Array, " K"]:
        return np.asarray(self._x)

    def y(self) -> Float[Array, " K"]:
        return np.asarray(self._y)

    def heading(self) -> Float[Array, " K"]:
        return np.asarray(self._heading)

    def positions(self) -> JaxObstacle2dPositionsForTimeStep:
        return JaxObstacle2dPositionsForTimeStep.create(x=self._x, y=self._y)

    def headings(self) -> JaxObstacleHeadingsForTimeStep:
        return JaxObstacleHeadingsForTimeStep.create(heading=self._heading)

    def replicate(self, *, horizon: int) -> JaxObstacle2dPoses:
        return JaxObstacle2dPoses.create(
            x=jnp.tile(self._x[jnp.newaxis, :], (horizon, 1)),
            y=jnp.tile(self._y[jnp.newaxis, :], (horizon, 1)),
            heading=jnp.tile(self._heading[jnp.newaxis, :], (horizon, 1)),
        )

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[0]

    @property
    def x_array(self) -> Float[JaxArray, " K"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, " K"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, " K"]:
        return self._heading

    @property
    def array(self) -> Float[JaxArray, f"{D_O} K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, f"{D_O} K"]:
        return jnp.stack([self._x, self._y, self._heading], axis=0)

    @cached_property
    def _numpy_array(self) -> Float[Array, "D_o K"]:
        return np.asarray(self._array)



from typing import Protocol, Sequence


class CostSumFunction[CostsT](Protocol):
    def __call__(self, costs: Sequence[CostsT], *, initial: CostsT) -> CostsT:
        """Sums an iterable of costs into a single cost."""
        ...
